#ifdef WIN32
#  pragma warning (disable: 4786)
#  include <windows.h>
#endif
#include "DFontGL.h"
#include <GL/GL.h>

bool DFontGL::stateSet = false;
bool DFontGL::glTexture2DVal;
bool DFontGL::glBlendVal;
int DFontGL::glBlendSrcVal;
int DFontGL::glBlendDstVal;
int DFontGL::screenWidth;
int DFontGL::screenHeight;

//////////////////////////////////////////////////////////////////////
// returns the nearest superior power of two
int ceilPow2(int n) {
  int e = 1;

  while (n) {
    n >>= 1;
    e <<= 1;
  }

  return e;
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
DFontGL::texture::texture() : 
  surface(NULL), 
  id(0)
  { }

//////////////////////////////////////////////////////////////////////
DFontGL::texture::~texture() { 
  if (surface != NULL) {
    SDL_FreeSurface(surface);
  }
  if (id) {
    glDeleteTextures(1, &id);
  }
}

//////////////////////////////////////////////////////////////////////
void DFontGL::texture::alloc(int w, int h) {
  Uint32 rmask, gmask, bmask, amask;

  /* SDL interprets each pixel as a 32-bit number, so our masks must depend
     on the endianness (byte order) of the machine */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
  rmask = 0xff000000;
  gmask = 0x00ff0000;
  bmask = 0x0000ff00;
  amask = 0x000000ff;
#else
  rmask = 0x000000ff;
  gmask = 0x0000ff00;
  bmask = 0x00ff0000;
  amask = 0xff000000;
#endif

  surface = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 32, rmask, gmask, bmask, amask);
}

//////////////////////////////////////////////////////////////////////
void DFontGL::texture::build() {
  glGenTextures(1, &id);
  glBindTexture(GL_TEXTURE_2D, id);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, surface->w, surface->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, surface->pixels);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP); 
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
}

//////////////////////////////////////////////////////////////////////
void DFontGL::texture::bind() {
  glBindTexture(GL_TEXTURE_2D, id);
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
DFontGL::character::character() {
}

//////////////////////////////////////////////////////////////////////
void DFontGL::character::set(texture *t_i, float u1_i, float v1_i, float u2_i, float v2_i, float w_i, float h_i) {
  t = t_i;
  u1 = u1_i;
  v1 = v1_i;
  u2 = u2_i;
  v2 = v2_i;
  w = w_i;
  h = h_i;
}

//////////////////////////////////////////////////////////////////////
void DFontGL::character::render(float x, float y) {
  t->bind();
  glBegin(GL_QUADS);
  glTexCoord2f(u1, v1);
  glVertex2f(x, y);
  glTexCoord2f(u2, v1);
  glVertex2f(x + w, y);
  glTexCoord2f(u2, v2);
  glVertex2f(x + w, y + h);
  glTexCoord2f(u1, v2);
  glVertex2f(x, y + h);
  glEnd();

  /*glBegin(GL_LINE_LOOP);
  glVertex2f(x, y);
  glVertex2f(x + w, y);
  glVertex2f(x + w, y + h);
  glVertex2f(x, y + h);
  glEnd();*/
}

//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////
DFontGL::DFontGL(const char * IFile, const char * DFile, int screenWidth_i, int screenHeight_i) :
  DFont(IFile, DFile),
  maxTextureWidth(256),
  maxTextureHeight(256)  
{
  screenWidth = screenWidth_i;
  screenHeight = screenHeight_i;
  makeTextures();
  makeMap();
}

//////////////////////////////////////////////////////////////////////
DFontGL::DFontGL() :
  DFont(),
  maxTextureWidth(256),
  maxTextureHeight(256)  
  { }

//////////////////////////////////////////////////////////////////////
void DFontGL::load(const char *file) {
  DFont::load(file);
  makeTextures();
  makeMap();
}

//////////////////////////////////////////////////////////////////////
void DFontGL::setScreenSize(int screenWidth_i, int screenHeight_i) {
  screenWidth = screenWidth_i;
  screenHeight = screenHeight_i;
}

//////////////////////////////////////////////////////////////////////
DFontGL::~DFontGL() {
}

//////////////////////////////////////////////////////////////////////
void DFontGL::buildTextures() {
  for (int i = 0 ; i < textures.size() ; i++) {
    textures[i].build();
  }
}

//////////////////////////////////////////////////////////////////////
void DFontGL::makeTextures() {
  int nbX, i;
  SDL_Rect srcRect;
  SDL_Rect dstRect;
  int textW, textH;
  int width;
  map<char, SDL_Rect>::const_iterator j;
  int fontSurfaceStart;
  Uint32 colorkey;
  Uint8 alpha;
  Uint32 flags;

  // verify that fontSurface has an usable format and set its alpha
  if (fontSurface->h > maxTextureHeight) {
    cerr << "DFontGL: font height must not exceed " << maxTextureHeight << " pixels" << endl;
    abort();
  }
  if (fontSurface->format->BytesPerPixel == 4) {
    // disable colorkeying and make surface opaque to keep image's alpha information
    // during contruction of textures
    colorkey = fontSurface->format->colorkey;
    alpha = fontSurface->format->alpha;
    flags = fontSurface->flags;
    SDL_SetColorKey(fontSurface, 0, 0);
    SDL_SetAlpha(fontSurface, 0, SDL_ALPHA_OPAQUE);
  }

  // calculate the number of required textures
  nbX = 1;
  fontSurfaceStart = 0;
  width = 0;
  for (j = fonts.begin() ; j != fonts.end() ; j++) {
    if (j->second.x + j->second.w > maxTextureWidth + fontSurfaceStart) {
      nbX++;
      fontSurfaceStart += width;
    } else {
      width = j->second.x + j->second.w - fontSurfaceStart;
    }
  }

  dstRect.x = 0;
  dstRect.y = 0;
  srcRect.x = 0;
  srcRect.y = 0;
  srcRect.h = fontSurface->h;
  textH = ceilPow2(fontSurface->h);
  j = fonts.begin();
  textures.reserve(nbX);
  fontSurfaceStart = 0;

  for (i = 0 ; i < nbX ; i++) {
    if (i == nbX - 1) {
      textW = ceilPow2(fontSurface->w - fontSurfaceStart);
    } else {
      textW = maxTextureWidth;
    }

    // get chars that fit in textW
    for (width = 0 ; j != fonts.end() ; j++) {
      if (j->second.x + j->second.w > textW + fontSurfaceStart)
        break;
      width = j->second.x + j->second.w - fontSurfaceStart;
    }

    srcRect.w = width;

    textures.push_back(texture());
    textures[i].alloc(textW, textH);
    textures[i].fontSurfaceStart = srcRect.x;
    SDL_BlitSurface(fontSurface, &srcRect, textures[i].surface, &dstRect);
    textures[i].build();

    srcRect.x += width;
    fontSurfaceStart = srcRect.x;
  }

  if (fontSurface->format->BytesPerPixel == 4) {
    // set back fontSurface in its initial state
    SDL_SetColorKey(fontSurface, flags & (SDL_RLEACCEL | SDL_SRCCOLORKEY), colorkey);
    SDL_SetAlpha(fontSurface, flags & (SDL_RLEACCEL | SDL_SRCALPHA), alpha);
  }
}

//////////////////////////////////////////////////////////////////////
void DFontGL::makeMap() {
  int tX;
  map<char, SDL_Rect>::const_iterator i;
  float u1, v1, u2, v2;

  for (i = fonts.begin() ; i != fonts.end() ; i++) {
    tX = getTexture(i->second.x);
    getUV(tX, i->second.x, i->second.y, u1, v1);
    getUV(tX, i->second.x + i->second.w, i->second.y + i->second.h - 1, u2, v2);
    _fonts[i->first] = character();
    _fonts[i->first].set(&textures[tX], u1, v1, u2, v2, i->second.w, i->second.h);
  }  
}

//////////////////////////////////////////////////////////////////////
int DFontGL::getTexture(int x) {
  int num = textures.size() - 1;

  while ((textures[num].fontSurfaceStart > x) && num) {
    num--;
  }

  return num;
}

//////////////////////////////////////////////////////////////////////
void DFontGL::prepareState() {
  if (!stateSet) {
    // setup a 2D transformation
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();
    glScalef(2.0 / (float)screenWidth, -2.0 / (float)screenHeight, 1.0);
    glTranslatef(-(float)screenWidth * 0.5, -(float)(screenHeight - 1) * 0.5, 0);

    glTexture2DVal = glIsEnabled(GL_TEXTURE_2D);
    glBlendVal = glIsEnabled(GL_BLEND);
    glGetIntegerv(GL_BLEND_SRC, &glBlendSrcVal);
    glGetIntegerv(GL_BLEND_DST, &glBlendDstVal);

    glEnable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    stateSet = true;
  }
}

//////////////////////////////////////////////////////////////////////
void DFontGL::restoreState() {
  if (stateSet) {
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();

    if (!glTexture2DVal)
      glDisable(GL_TEXTURE_2D);
    if (!glBlendVal)
      glDisable(GL_BLEND);
    else {
      glBlendFunc(glBlendSrcVal, glBlendDstVal);
    }
    
    stateSet = false;
  }
}


//////////////////////////////////////////////////////////////////////
void DFontGL::PutString(Sint16 x, Sint16 y, const char *str) {
  bool changedState;

  if (!stateSet) {
    prepareState();
    changedState = true;
  } else {
    changedState = false;
  }

  if (x == CTR_HORIZ) {
    int stringWidth = 0;
    char const * ptr = str;
    for (; *ptr != '\0'; ++ptr) {
      if (*ptr == ' ')
        stringWidth += gap;
      else
        stringWidth += _fonts[*ptr].w;
    }
    x = (screenWidth / 2) - (stringWidth / 2);
  }

  if (y == CTR_VERT)
    y = (screenHeight / 2) - (_fonts['!'].h / 2);
  
  for (; *str != '\0'; ++str) {
    if (*str == ' ') 
      x += gap;  
    else {
      if (_fonts.count(*str)) {
        _fonts[*str].render(x, y);      
        x += _fonts[*str].w;
      } else {
        x += gap;
      }
    }
  }

  if (changedState) {
    restoreState();
  }
}

//////////////////////////////////////////////////////////////////////
void DFontGL::PutChar(Sint16 x, Sint16 y, const char c) {
  bool changedState;

  if (!stateSet) {
    prepareState();
    changedState = true;
  } else {
    changedState = false;
  }

  if (_fonts.count(c)) {
    _fonts[c].render(x, y);
  }

  if (changedState) {
    restoreState();
  }
}

//////////////////////////////////////////////////////////////////////
void DFontGL::PutInt (Sint16 x, Sint16 y, int num) {
	char str[40]; //more than enough space.
	sprintf(str, "%d", num);
	PutString(x, y, str);
}

//////////////////////////////////////////////////////////////////////
void DFontGL::getUV(int textX, int x, int y, float &u, float &v) {
  u = (float)(x - textures[textX].fontSurfaceStart) / (float)textures[textX].surface->w;
  v = (float)y / (float)textures[textX].surface->h;
}